; EMS routines for SMALLOC.

.MODEL						SMALL

extrn						DOS3CALL:far

TRUE						equ 1
FALSE						equ 0

.DATA

	EMMName					db 'EMMXXXX0',0

.CODE

public						_EMSCheckPresence

_EMSCheckPresence			proc near

	; Returns TRUE if EMM is present, FALSE if not.
	; Must use OPEN method, as getting the interrupt address and adding
	; 10 to the offset to look for EMMXXXX0 doesn't work in protected modes.

							mov dx,offset EMMName
							mov ax,03D00h
							call DOS3CALL ; Try to open
							jc EMSCP_1

							mov bx,ax
							mov ax,04400h
							call DOS3CALL ; IOCTL Get device info
							jc EMSCP_2
							test dx,080h ; Check it's a char dev
							jz EMSCP_2

							mov ax,04407h
							call DOS3CALL ; IOCTL get output status
							jc EMSCP_2
							cmp al,0 ; If al=0, EMM not there
							je EMSCP_2

	EMSCP_2:
							mov ah,03Eh
							call DOS3CALL ; Close
							jc EMSCP_1

							mov ax,TRUE
							jmp EMSCP_3

	EMSCP_1:				mov ax,FALSE

	EMSCP_3:				ret

_EMSCheckPresence			endp

public						_EMSGetStatus

_EMSGetStatus				proc near

	; Returns EMM error code (0=OK).
							mov ah,040h
							int 067h
							mov al,ah
							xor ah,ah
							ret

_EMSGetStatus				endp

public						_EMSGetPageFrameAddress

_EMSGetPageFrameAddress		proc near

	; Returns EMM error code (0=OK). If no error, *lpwVersion [bp+4]
	; has the selector (Windows in protected mode works OK too) address.

							push bp
							mov bp,sp
							push di

							mov ah,041h
							int 067h
							cmp ah,0
							jne EMSGPFA_1
							les di,[bp+4]
							mov es:[di],bx

	EMSGPFA_1:				mov al,ah
							xor ah,ah

							pop di
							pop bp
							ret

_EMSGetPageFrameAddress		endp

public						_EMSAllocateHandleAndPages

_EMSAllocateHandleAndPages	proc near

	; Returns EMM error code (0=OK). If no error, *lpwHandle [bp+6]
	; has the EMM handle. wNumPages [bp+4] has number of pages to allocate.

							push bp
							mov bp,sp

							mov bx,[bp+4]
							mov ah,043h
							int 067h
							cmp ah,0
							jne EMSAHAP_1
							les bx,[bp+6]
							mov es:[bx],dx

	EMSAHAP_1:				mov al,ah
							xor ah,ah

							pop bp
							ret

_EMSAllocateHandleAndPages	endp

public						_EMSMapExpandedMemoryPage

_EMSMapExpandedMemoryPage	proc near

	; Returns EMM error code (0=OK). wHandle [bp+4] is the EMM Handle,
	; cPhysicalPage [bp+6] is the physical page and wLogicalPage [bp+8]
	; is the logical page to map.

							push bp
							mov bp,sp

							mov dx,[bp+4]
							mov al,[bp+6]
							mov bx,[bp+8]
							mov ah,044h
							int 067h
							mov al,ah
							xor ah,ah

							pop bp
							ret

_EMSMapExpandedMemoryPage	endp

public						_EMSReleaseHandleAndMemory

_EMSReleaseHandleAndMemory	proc near

	; Returns EMM error code (0=OK). wHandle [bp+4] is the EMM Handle to
	; release.

							push bp
							mov bp,sp

							mov dx,[bp+4]
							mov ah,045h
							int 067h
							mov al,ah
							xor ah,ah

							pop bp
							ret

_EMSReleaseHandleAndMemory	endp

public						_EMSGetVersion

_EMSGetVersion				proc near

	; Returns EMM error code (0=OK). If no error, *lpcVersion [bp+4]
	; has the EMM version BCD encoded (hi nybble=major, lo nybble=minor).

							push bp
							mov bp,sp

							mov ah,046h
							int 067h
							cmp ah,0
							jne EMSGV_1
							les bx,[bp+4]
							mov es:[bx],al

	EMSGV_1:				mov al,ah
							xor ah,ah

							pop bp
							ret

_EMSGetVersion				endp

public						_EMSSavePageMap

_EMSSavePageMap				proc near

	; Returns EMM error code (0=OK). *lpMap [bp+4] receives the mapping
	; information

							push bp
							mov bp,sp

							push di

							mov ah,04Eh
							mov al,000h
							les di,[bp+4]
							int 067h
							cmp ah,0
							jne EMSSPM_1

	EMSSPM_1:				mov al,ah
							xor ah,ah

							pop di
							pop bp

							ret

_EMSSavePageMap				endp

public						_EMSRestorePageMap

_EMSRestorePageMap			proc near

	; Returns EMM error code (0=OK). *lpMap [bp+4] holds the new mapping
	; information

							push bp
							mov bp,sp

							push si

							mov ah,04Eh
							mov al,001h
							push ds
							lds si,[bp+4]
							int 067h
							pop ds
							cmp ah,0
							jne EMSRPM_1

	EMSRPM_1:				mov al,ah
							xor ah,ah

							pop si
							pop bp

							ret

_EMSRestorePageMap			endp

public						_EMSSaveAndRestorePageMap

_EMSSaveAndRestorePageMap	proc near

	; Returns EMM error code (0=OK). *lpNewMap [bp+4] is the new mapping
	; information, and *lpOldMap [bp+8] is the old mapping info to receive
	; data.

							push bp
							mov bp,sp

							push di
							push si

							mov ah,04Eh
							mov al,002h
							push ds
							lds si,[bp+4]
							les di,[bp+8]
							int 067h
							pop ds
							cmp ah,0
							jne EMSSARPM_1

	EMSSARPM_1:				mov al,ah
							xor ah,ah

							pop si
							pop di
							pop bp

							ret

_EMSSaveAndRestorePageMap	endp

public						_EMSGetSizeOfPageMapInfo

_EMSGetSizeOfPageMapInfo	proc near

	; Returns EMM error code (0=OK). If no error, *lpcSize [bp+4]
	; has the size of the buffer required to save page mapping info.

							push bp
							mov bp,sp

							mov ah,04Eh
							mov al,003h
							int 067h
							cmp ah,0
							jne EMSGSOPMI_1
							les bx,[bp+4]
							mov es:[bx],al

	EMSGSOPMI_1:			mov al,ah
							xor ah,ah

							pop bp

							ret

_EMSGetSizeOfPageMapInfo	endp

public						_EMSReallocateHandlePages

_EMSReallocateHandlePages	proc near

	; Returns EMM error code (0=OK). wHandle [bp+4] is the EMM handle,
	; wSize [bp+6] is the number of pages.

							push bp
							mov bp,sp

							mov dx,[bp+4]
							mov bx,[bp+6]
							mov ah,051h
							int 067h
							mov al,ah
							xor ah,ah

							pop bp
							ret

_EMSReallocateHandlePages	endp

end
