/*

Memory allocation routines for large numbers of small objects.

Author: Howard Long

Date started: 31 August 1991

Rel 0.1: 15 September 1991: Basic allocation functionality completed.

Rel 1.0: 1 October 1991: Virtual memory support added.

*/

#define _WINDOWS
#define _WINDLL
#define NOCOMM
#include <windows.h>

#ifdef DEBUG
	#include <string.h>
#endif

/* For _fmemcpy()... */
#ifdef __TURBOC__
	#include <mem.h>
#else
	#include <memory.h>
#endif

#include "smalloc.h"
#include "emsasm.h"
#include "fileasm.h"

/* Undocumented or partly documented Windows functions. */
extern DWORD FAR PASCAL GetSelectorBase(WORD wSelector);
extern DWORD FAR PASCAL GetSelectorLimit(WORD wSelector);
extern int FAR PASCAL DOS3Call(void);

/* Not defined in MS Windows SDK 3.0 WINDOWS.H but is in Borland's. */
#ifndef __TURBOC__
	extern DWORD FAR PASCAL GlobalDosAlloc(DWORD);
	extern WORD FAR PASCAL GlobalDosFree(WORD);
#endif

#ifdef DEBUG
	#define DEBUGERROR(smee) { SMDebug((smee),__LINE__); }
	#define DEBUGWARNING(smee) { SMWarning((smee),__LINE__); }
#else
	#define DEBUGERROR(smee) {smee+=0;}
	#define DEBUGWARNING(smee) {smee+=0;}
#endif

#define ARENAHEADERSTRUCTSIZE (sizeof(ARENAHEADERSTRUCT))
#define BLOCKSTRUCTSIZE (sizeof(BLOCKSTRUCT))
#define TYPESTRUCTSIZE (sizeof(TYPESTRUCT)-BLOCKSTRUCTSIZE)
#define FREESTRUCTSIZE (sizeof(FREESTRUCT))
#define NULLBLOCKSTRUCTINDEX (0xFFFF)
#define NULLEMSHANDLE (0xFFFF)

/* Max sizes before GlobalReAlloc can change handle. */
#define MAXENHANCED (65536UL)
#define MAXSTANDARD (65519UL)

/* Size of copy buffer */
#define BUFFERSIZE (4096)

typedef VOID NEAR *NPVOID;

typedef enum {STE_NORM,STE_EMS,STE_FILE} SWAPTYPEENUM;

/* ARENAHEADERSTRUCT appears at the head of every ghData allocation */
/* so that the NotifyProc may identify it. */
typedef struct
{
	LOCALHANDLE lhts; /* Owner TYPESTRUCT handle. */
	WORD wbsIndex; /* bs Index. */
} ARENAHEADERSTRUCT;
typedef ARENAHEADERSTRUCT FAR *LPARENAHEADERSTRUCT;

typedef struct
{
	GLOBALHANDLE ghData; /* Global handle to data. NULL if not used. */
	WORD wfsNum; /* Number of free list items. */
	WORD wfsLargestIndex; /* Index to largest free entry. */
	WORD wfsLargestSize; /* Size of largest free entry. */
	WORD wLockCount; /* Indicates if data is locked. */
	WORD wSwapPos; /* Position in secondary memory of data. */
	SWAPTYPEENUM ste; /* Type of swap. */
	union
	{
		struct
		{
			WORD wDataSeg; /* NULL if unlocked, otherwise the */
						   /* segment/selector of ghData. */
			WORD wDataNum; /* Number of data items. */
		} Used;
		struct
		{
			WORD wbsFreeNext; /* Next free bs index. */
			WORD wbsFreePrev; /* Previous free bs index. */
		} Free;
	} Union;
} BLOCKSTRUCT;
typedef BLOCKSTRUCT NEAR *NPBLOCKSTRUCT;

typedef struct
{
	WORD wbsNum; /* Number of BLOCKSTRUCTs assigned to type. */
	WORD wDataSize; /* Size of each ghData. */
	WORD wfsMin; /* Minimum number of free list entries for each block. */
	WORD wbsFreeIndex; /* Index to next free bs, or 0 if non free. */
	WORD wEMSHandle; /* EMSHandle for notification. NULL if not used. */
	WORD wEMSNum; /* Number of EMS allocations. */
	WORD wEMSbsIndex; /* Index to bs's swapped to EMS. */
	WORD wFlags; /* GlobalAlloc flags. */
	WORD wFilebsIndex; /* Index to bs's swapped to file. */
	WORD wFileNum; /* Number of file allocations. */
	WORD wSelBuffer; /* Selector of buffer in lower 1Mb for prot modes. */
	WORD wbsLastAllocIndex; /* Index to last bs allocated: used for speed. */
	char szPathName[144]; /* Path name for swapper. */
#ifdef ALWAYSOPEN
	int nfh; /* File handle. */
	WORD wPSP; /* Owner's PSP. */
#endif
	BLOCKSTRUCT abs[1]; /* Array of blocks, wbsNum elements in size. */
} TYPESTRUCT;
typedef TYPESTRUCT NEAR *NPTYPESTRUCT;

typedef struct
{
	/* This structure should be a power of 2 in size for HUGE pointers. */
	WORD wOffStart; /* Start offset of free list entry. */
	WORD wOffEnd; /* End offset of free list entry. */
} FREESTRUCT;
typedef FREESTRUCT FAR *LPFREESTRUCT;

static DWORD _dwWinFlags;
static BOOL _bEMSPresent=FALSE;
static LOCALHANDLE _lhPageMapInfo=NULL;
static LOCALHANDLE _lhBuffer=NULL; /* Real mode buffer. */
static LPSTR _lpsEMSPageFrame=NULL; /* EMS Page frame selector base. */
static GLOBALHANDLE _ghNoDiscard=NULL; /* If not NULL, NotifyProc won't */
									   /* discard. */

#ifdef DEBUG

static VOID litoa(LPSTR lpsz,WORD w)
{
	LPSTR lpszTemp=lpsz+5;

	*lpszTemp='\0';
	while (lpszTemp--!=lpsz)
	{
		*lpszTemp=(char)(int)(w%10)+'0';
		w/=10;
	}
}

VOID SMDebug(SMERRORENUM smee,WORD wLine)
{
	char sz[40];
	char sz2[7];

	_fstrcpy(sz,"SM Error #");
	litoa(sz2,smee);
	_fstrcat(sz,sz2);
	_fstrcat(sz," @ line ");
	litoa(sz2,wLine);
	_fstrcat(sz,sz2);
	_fstrcat(sz,"\r\n");

	OutputDebugString(sz);
}

VOID SMWarning(SMERRORENUM smee,WORD wLine)
{
	char sz[40];
	char sz2[7];

	_fstrcpy(sz,"SM Warning #");
	litoa(sz2,smee);
	_fstrcat(sz,sz2);
	_fstrcat(sz," @ line ");
	litoa(sz2,wLine);
	_fstrcat(sz,sz2);
	_fstrcat(sz,"\r\n");

	OutputDebugString(sz);
}
#endif

BOOL FAR PASCAL LibMain(HANDLE hInstance,WORD wDataSeg,WORD wHeapSize,
	LPSTR lpszCmdLine)
{
	/* Stop nasty warnings... */
	hInstance+=0;
	wDataSeg+=0;
	lpszCmdLine+=0;

	if (wHeapSize==0)
	{
		return FALSE;
	}

	/* Unlock data segment because in real mode large frame things can get */
	/* sticky in the non-banked area. Note that you musn't call any */
	/* functions that may move memory if you make a local pointer into */
	/* a far pointer. */

	UnlockData(0);

	_dwWinFlags=GetWinFlags();

	/* Check if buffer required in first 1Mb... */
	if ((_dwWinFlags & WF_PMODE)==0)
	{
		/* Need to allocate a buffer for real mode. */
		if ((_lhBuffer=LocalAlloc(LMEM_MOVEABLE,BUFFERSIZE))==NULL)
		{
			return FALSE;
		}
	}

	/* Check if EMS is OK... */
	{
		BYTE cSize;
		BYTE cVersion;
		WORD wPageFrameSel;

		if (EMSCheckPresence() && EMSGetStatus()==SMEE_OK &&
			EMSGetSizeOfPageMapInfo(&cSize)==SMEE_OK &&
			EMSGetVersion(&cVersion)==SMEE_OK &&
			cVersion>=0x40 &&
			EMSGetPageFrameAddress(&wPageFrameSel)==SMEE_OK &&
			(_lhPageMapInfo=LocalAlloc(LMEM_MOVEABLE,(WORD)cSize))!=NULL)
		{
			_lpsEMSPageFrame=(LPSTR)MAKELONG(0,wPageFrameSel);
			_bEMSPresent=TRUE;
		}
	}

	return TRUE;
}

int FAR PASCAL WEP(int nSystemExit)
{
	if (_lhPageMapInfo!=NULL)
	{
		LocalFree(_lhPageMapInfo);
		_lhPageMapInfo=NULL;
	}

	if (_lhBuffer!=NULL)
	{
		LocalFree(_lhBuffer);
		_lhBuffer=NULL;
	}

	switch (nSystemExit)
	{
		case WEP_SYSTEM_EXIT:
			break;
		case WEP_FREE_DLL:
			break;
	}
	return 1;
}

static DWORD SMPhysicalToEMS(DWORD dwPos)
{
	/* Returns HIWORD as the logical page and LOWORD as the offset. */
	return MAKELONG(dwPos & 0x3FFF,dwPos >> 14);
}

static SMERRORENUM SMEMSToEMS(WORD wEMSHandle,DWORD dwPosDst,DWORD dwPosSrc,
	WORD wSize)
{
	/* Copy EMS memory to EMS. */
	SMERRORENUM smee=SMEE_OK;
	NPVOID np=LocalLock(_lhPageMapInfo);

	if ((smee=EMSSavePageMap(np))!=SMEE_OK)
	{
		DEBUGERROR(smee);
	}
	else
	{
		DWORD dwEMSPosDst=SMPhysicalToEMS(dwPosDst);
		WORD wLogicalPageDst=HIWORD(dwEMSPosDst);
		WORD wEMSOffsetDst=LOWORD(dwEMSPosDst);
		DWORD dwEMSPosSrc=SMPhysicalToEMS(dwPosSrc);
		WORD wLogicalPageSrc=HIWORD(dwEMSPosSrc);
		WORD wEMSOffsetSrc=LOWORD(dwEMSPosSrc);
		WORD wNumToCopy=wSize;

		while (smee==SMEE_OK && wNumToCopy>0)
		{
			LPSTR lpsDst=_lpsEMSPageFrame+wEMSOffsetDst;
			LPSTR lpsSrc=_lpsEMSPageFrame+wEMSOffsetSrc;

			if ((smee=EMSMapExpandedMemoryPage(wEMSHandle,'\0',
				wLogicalPageDst))!=SMEE_OK)
			{
				DEBUGERROR(smee);
			}

			if (smee==SMEE_OK && wLogicalPageDst!=wLogicalPageSrc)
			{
				if ((smee=EMSMapExpandedMemoryPage(wEMSHandle,'\1',
					wLogicalPageSrc))!=SMEE_OK)
				{
					DEBUGERROR(smee);
				}
				lpsSrc+=16384; /* It's in the next page. */
			}

			if (smee==SMEE_OK)
			{
				WORD wXferSize=min(wNumToCopy,
					min(16384U-wEMSOffsetDst,16384U-wEMSOffsetSrc));
				_fmemcpy(lpsDst,lpsSrc,wXferSize);

				wEMSOffsetDst+=wXferSize;
				wEMSOffsetSrc+=wXferSize;

				if (wEMSOffsetDst==16384)
				{
					wEMSOffsetDst=0;
					wLogicalPageDst++;
				}
				if (wEMSOffsetSrc==16384)
				{
					wEMSOffsetSrc=0;
					wLogicalPageSrc++;
				}
				wNumToCopy-=wXferSize;
			}
		}
		EMSRestorePageMap(np);
	}
	LocalUnlock(_lhPageMapInfo);
	np=NULL;
	return smee;
}

static SMERRORENUM SMNormToEMS(WORD wEMSHandle,DWORD dwPos,LPSTR lps,
	WORD wSize)
{
	/* Copy normal memory to EMS. */
	SMERRORENUM smee=SMEE_OK;
	NPVOID np=LocalLock(_lhPageMapInfo);

	if ((smee=EMSSavePageMap(np))!=SMEE_OK)
	{
		DEBUGERROR(smee);
	}
	else
	{
		DWORD dwEMSPos=SMPhysicalToEMS(dwPos);
		WORD wLogicalPage=HIWORD(dwEMSPos);
		WORD wEMSOffset=LOWORD(dwEMSPos);
		WORD wNumToCopy=wSize;

		while (smee==SMEE_OK && wNumToCopy>0)
		{
			if ((smee=
				EMSMapExpandedMemoryPage(wEMSHandle,'\0',wLogicalPage))!=
				SMEE_OK)
			{
				DEBUGERROR(smee);
			}
			else
			{
				WORD wXferSize=min(wNumToCopy,16384U-wEMSOffset);

				_fmemcpy(_lpsEMSPageFrame+wEMSOffset,lps,wXferSize);

				wEMSOffset=0;
				wLogicalPage++;
				wNumToCopy-=wXferSize;
				lps+=wXferSize;
			}
		}

		if ((smee=EMSRestorePageMap(np))!=SMEE_OK)
		{
			DEBUGERROR(smee);
		}
	}
	LocalUnlock(_lhPageMapInfo);
	np=NULL;
	return smee;
}

static SMERRORENUM SMEMSToNorm(LPSTR lps,WORD wEMSHandle,DWORD dwPos,WORD wSize)
{
	/* Copy EMS memory to normal. */
	SMERRORENUM smee=SMEE_OK;
	NPVOID np=LocalLock(_lhPageMapInfo);

	if ((smee=EMSSavePageMap(np))!=SMEE_OK)
	{
		DEBUGERROR(smee);
	}
	else
	{
		DWORD dwEMSPos=SMPhysicalToEMS(dwPos);
		WORD wLogicalPage=HIWORD(dwEMSPos);
		WORD wEMSOffset=LOWORD(dwEMSPos);
		WORD wNumToCopy=wSize;

		while (smee==SMEE_OK && wNumToCopy>0)
		{
			if ((smee=
				EMSMapExpandedMemoryPage(wEMSHandle,'\0',wLogicalPage))!=
				SMEE_OK)
			{
				DEBUGERROR(smee);
			}
			else
			{
				WORD wXferSize=min(wNumToCopy,16384U-wEMSOffset);

				_fmemcpy(lps,_lpsEMSPageFrame+wEMSOffset,wXferSize);

				wEMSOffset=0;
				wLogicalPage++;
				wNumToCopy-=wXferSize;
				lps+=wXferSize;
			}
		}

		EMSRestorePageMap(np);
	}
	LocalUnlock(_lhPageMapInfo);
	np=NULL;
	return smee;
}

static SMERRORENUM SMFileToFile(int nfh,DWORD dwPosDst,DWORD dwPosSrc,
	WORD wSize,WORD wSelBuffer)
{
	/* Copy file memory to file. */
	SMERRORENUM smee=SMEE_OK;
	WORD wTotalToXfer=wSize;
	LPSTR lpsBuffer=NULL;

	if ((_dwWinFlags & WF_PMODE)!=0)
	{
		lpsBuffer=(LPSTR)MAKELONG(0,wSelBuffer);
	}
	else
	{
		lpsBuffer=(LPSTR)LocalLock(_lhBuffer);
	}

	while (smee==SMEE_OK && wTotalToXfer>0)
	{
		WORD wNumToXfer=min(BUFFERSIZE,wTotalToXfer);
		DWORD dwPos=0UL;

		if ((smee=FileSeek(nfh,dwPosSrc,0,&dwPos))!=SMEE_OK)
		{
			DEBUGERROR(smee);
		}
		else
		{
			if (dwPos!=dwPosSrc)
			{
				smee=SMEE_FILESEEKERROR;
				DEBUGERROR(smee);
			}
		}

		if (smee==SMEE_OK)
		{
			WORD wNumRead=0U;

			if ((smee=FileRead(nfh,lpsBuffer,wNumToXfer,&wNumRead))!=SMEE_OK)
			{
				DEBUGERROR(smee);
			}
			else
			{
				if (wNumRead!=wNumToXfer)
				{
					smee=SMEE_FILEREADERROR;
					DEBUGERROR(smee);
				}
			}
		}

		if (smee==SMEE_OK)
		{
			DWORD dwPos=0UL;

			if ((smee=FileSeek(nfh,dwPosDst,0,&dwPos))!=SMEE_OK)
			{
				DEBUGERROR(smee);
			}
			else
			{
				if (dwPos!=dwPosDst)
				{
					smee=SMEE_FILESEEKERROR;
					DEBUGERROR(smee);
				}
			}
		}

		if (smee==SMEE_OK)
		{
			WORD wNumWritten=0U;

			if ((smee=FileWrite(nfh,lpsBuffer,wNumToXfer,&wNumWritten))!=
				SMEE_OK)
			{
				DEBUGERROR(smee);
			}
			else
			{
				if (wNumWritten!=wNumToXfer)
				{
					smee=SMEE_FILEWRITEERROR;
					DEBUGERROR(smee);
				}
			}
		}

		dwPosDst+=wNumToXfer;
		dwPosSrc+=wNumToXfer;
		wTotalToXfer-=wNumToXfer;
	}

	if ((_dwWinFlags & WF_PMODE)==0)
	{
		LocalUnlock(_lhBuffer);
		lpsBuffer=NULL;
	}

	return smee;
}

static SMERRORENUM SMNormToFile(int nfh,DWORD dwPos,LPSTR lps,WORD wSize,
	WORD wSelBuffer)
{
	/* Copy normal memory to file. */
	SMERRORENUM smee=SMEE_OK;
	DWORD dwSeekPos=0UL;

	if ((smee=FileSeek(nfh,dwPos,0,&dwSeekPos))!=SMEE_OK)
	{
		DEBUGERROR(smee);
	}
	else
	{
		if (dwSeekPos!=dwPos)
		{
			smee=SMEE_FILESEEKERROR;
			DEBUGERROR(smee);
		}
	}

	/* Check to see if lps totally lies physically within 1Mb range for */
	/* DOS. */
	if (smee==SMEE_OK)
	{
		if ((_dwWinFlags & WF_PMODE)==0UL ||
			GetSelectorBase(HIWORD((DWORD)lps))+
			GetSelectorLimit(HIWORD((DWORD)lps)) < 0x100000UL)
		{
			/* Below 1Mb, so all's OK... */
			WORD wNumWritten=0U;

			if ((smee=FileWrite(nfh,lps,wSize,&wNumWritten))!=SMEE_OK)
			{
				DEBUGERROR(smee);
			}
			else
			{
				if (wNumWritten!=wSize)
				{
					smee=SMEE_FILEWRITEERROR;
					DEBUGERROR(smee);
				}
			}
		}
		else
		{
			/* Got to use our DOS buffer to perform writes... */
			WORD wTotalToWrite=wSize;
			LPSTR lpsBuffer=(LPSTR)MAKELONG(0,wSelBuffer);

			while (smee==SMEE_OK && wTotalToWrite>0)
			{
				WORD wNumToWrite=min(BUFFERSIZE,wTotalToWrite);
				WORD wNumWritten=0U;

				_fmemcpy(lpsBuffer,lps,wNumToWrite);

				if ((smee=FileWrite(nfh,lps,wNumToWrite,&wNumWritten))!=
					SMEE_OK)
				{
					DEBUGERROR(smee);
				}
				else
				{
					if (wNumWritten!=wNumToWrite)
					{
						smee=SMEE_FILEWRITEERROR;
						DEBUGERROR(smee);
					}
				}
				wTotalToWrite-=wNumToWrite;
				lps+=wNumToWrite;
			}
		}
	}
	return smee;
}

static SMERRORENUM SMFileToNorm(LPSTR lps,int nfh,DWORD dwPos,WORD wSize,
	WORD wSelBuffer)
{
	/* Copy file memory to normal. */
	SMERRORENUM smee=SMEE_OK;
	DWORD dwSeekPos=0UL;

	if ((smee=FileSeek(nfh,dwPos,(BYTE)0,&dwSeekPos))!=SMEE_OK)
	{
		DEBUGERROR(smee);
	}
	else
	{
		if (dwPos!=dwSeekPos)
		{
			smee=SMEE_FILESEEKERROR;
			DEBUGERROR(smee);
		}
	}

	/* Check to see if lps totally lies physically within 1Mb range for */
	/* DOS. */
	if (smee==SMEE_OK)
	{
		if ((_dwWinFlags & WF_PMODE)==0 ||
			GetSelectorBase(HIWORD((DWORD)lps))+
			GetSelectorLimit(HIWORD((DWORD)lps)) < 0x100000UL)
		{
			/* Below 1Mb, so all's OK... */
			WORD wNumRead=0U;

			if ((smee=FileRead(nfh,lps,wSize,&wNumRead))!=SMEE_OK)
			{
				DEBUGERROR(smee);
			}
			else
			{
				if (wNumRead!=wSize)
				{
					smee=SMEE_FILEREADERROR;
					DEBUGERROR(smee);
				}
			}
		}
		else
		{
			/* Got to use our DOS buffer to perform reads... */
			WORD wTotalToRead=wSize;
			LPSTR lpsBuffer=(LPSTR)MAKELONG(0,wSelBuffer);

			while (smee==SMEE_OK && wTotalToRead>0)
			{
				WORD wNumToRead=min(BUFFERSIZE,wTotalToRead);
				WORD wNumRead=0U;

				if ((smee=FileRead(nfh,lpsBuffer,wNumToRead,&wNumRead))!=
					SMEE_OK)
				{
					DEBUGERROR(smee);
				}
				else
				{
					if (wNumRead!=wNumToRead)
					{
						smee=SMEE_FILEREADERROR;
						DEBUGERROR(smee);
					}
				}
				_fmemcpy(lps,lpsBuffer,wNumToRead);
				wTotalToRead-=wNumToRead;
				lps+=wNumToRead;
			}
		}
	}
	return smee;
}

BOOL FAR PASCAL NotifyProc(GLOBALHANDLE gh)
{
	SMERRORENUM smee=SMEE_OK;
	SWAPTYPEENUM ste=STE_NORM;

#ifdef DEBUG
	static WORD wReentry=0;

	if (wReentry>0)
	{
		OutputDebugString("********NotifyProc reentered.\r\n");
	}
	wReentry++;

	OutputDebugString("NotifyProc entered...\r\n");
#endif

	if (gh==_ghNoDiscard)
	{
		/* Don't discard & fail. */

#ifdef DEBUG
		OutputDebugString("NotifyProc can't discard...\r\n");
#endif

		ste=STE_NORM;
	}
	else
	{
		/* Who's the owner? */
		LPARENAHEADERSTRUCT lpahs=(LPARENAHEADERSTRUCT)GlobalLock(gh);
		LOCALHANDLE lhts=lpahs->lhts;
		WORD wbsIndex=lpahs->wbsIndex;
		NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhts);
		NPBLOCKSTRUCT npbsBase=npts->abs;
		NPBLOCKSTRUCT npbs=npbsBase+wbsIndex;
		WORD wEMSOldPages=(WORD)
			(((DWORD)npts->wDataSize*npts->wEMSNum+16383UL)/16384UL);
		WORD wEMSNewPages=(WORD)
			(((DWORD)npts->wDataSize*(npts->wEMSNum+1)+16383UL)/16384UL);

		/* Try first to get space in EMS... */
		if ((npts->wFlags & SMEM_SWAPEMS)!=0 &&
			_bEMSPresent && EMSGetStatus()==SMEE_OK)
		{
			if (npts->wEMSHandle==NULLEMSHANDLE)
			{
				/* Try to get an EMS handle... */
				if ((smee=EMSAllocateHandleAndPages(wEMSNewPages,
					&npts->wEMSHandle))!=SMEE_OK)
				{
					DEBUGWARNING(smee);
					smee=SMEE_OK;
				}
				else
				{
					ste=STE_EMS;
				}
			}
			else
			{
				/* Reallocate exiting handle if necessary... */
				if (wEMSNewPages==wEMSOldPages)
				{
					ste=STE_EMS;
				}
				else
				{
					if ((smee=EMSReallocateHandlePages(npts->wEMSHandle,
						wEMSNewPages))!=SMEE_OK)
					{
						DEBUGWARNING(smee);
						smee=SMEE_OK;
					}
					else
					{
						ste=STE_EMS;
					}
				}
			}
		}

		if (ste==STE_EMS)
		{
			/* Perform the Normal to EMS transfer... */
			WORD wSwapPos=npts->wEMSNum;
			DWORD dwSwapPos=(DWORD)npts->wDataSize*wSwapPos;

			if ((smee=SMNormToEMS(npts->wEMSHandle,dwSwapPos,(LPSTR)lpahs,
				npts->wDataSize))!=SMEE_OK)
			{
				DEBUGWARNING(smee);
				smee=SMEE_OK;

				if (wEMSOldPages==0)
				{
					/* Free EMS handle... */
					if ((smee=EMSReleaseHandleAndMemory(npts->wEMSHandle))!=
						SMEE_OK)
					{
						DEBUGWARNING(smee);
						smee=SMEE_OK;
					}
					npts->wEMSHandle=NULLEMSHANDLE;
				}
				else
				{
					/* Reallocate EMS handle back to old size... */
					if ((smee=EMSReallocateHandlePages(npts->wEMSHandle,
						wEMSOldPages))!=SMEE_OK)
					{
						DEBUGWARNING(smee);
						smee=SMEE_OK;
					}
				}
				ste=STE_NORM;
			}
			else
			{
				/* At last, EMS has succeeded... */
				npbs->wSwapPos=wSwapPos;
				npbs->ste=STE_EMS;
				npts->wEMSNum++;
			}
		}

		if (ste==STE_NORM && (npts->wFlags & SMEM_SWAPFILE)!=0 &&
			npts->szPathName[0]!='\0')
		{
			/* We'll need to give file a chance. */
#ifdef ALWAYSOPEN
			int nfh=npts->nfh;
			WORD wOwnerPSP=npts->wPSP;
			WORD wCurrentPSP;

			/* Get current PSP (cannot use GetCurrentPDB as it's in a */
			/* moveable discardable code segment). */
			{
				/* DOS 3.0+ function 0x63: Win 3 must have DOS 3, so it's */
				/* OK to know we can use this function. */
				_asm mov ah,062h
				_asm call DOS3Call
				_asm mov wCurrentPSP,bx
			}

			/* Set to proper PSP for file I/O... */
			if (wCurrentPSP!=wOwnerPSP)
			{
				/* Set PSP so we can use our file handle... */
				_asm mov bx,wOwnerPSP
				_asm mov ah,050h
				DOS3Call();
			}
#else
			int nfh=-1;

			if ((smee=FileOpen(npts->szPathName,
				OF_READWRITE | OF_SHARE_EXCLUSIVE,&nfh))!=SMEE_OK)
			{
				DEBUGWARNING(smee);
				smee=SMEE_OK;
			}
			else
#endif
			{
				/* Try to extend the size of the file... */
				WORD wSwapPos=npts->wFileNum;
				DWORD dwOldSize=(DWORD)npts->wDataSize*npts->wFileNum;
				DWORD dwNewSize=(DWORD)npts->wDataSize*(npts->wFileNum+1);
				DWORD dwPos;

				ste=STE_FILE;

				if ((smee=FileSeek(nfh,dwNewSize,0,&dwPos))!=SMEE_OK)
				{
					DEBUGWARNING(smee);
					smee=SMEE_OK;
					ste=STE_NORM;
				}

				if (dwPos!=dwNewSize)
				{
					smee=SMEE_FILESEEKERROR;
					DEBUGWARNING(smee);
					smee=SMEE_OK;
					ste=STE_NORM;
				}

				if (ste==STE_FILE)
				{
					/* Check we can write at end of file to fix size */
					/* (there may not be enough disk space)... */
					WORD wNumWritten;

					if ((smee=FileWrite(nfh,NULL,0,&wNumWritten))!=SMEE_OK)
					{
						DEBUGWARNING(smee);
						smee=SMEE_OK;
						ste=STE_NORM;
					}
					if (wNumWritten!=0)
					{
						smee=SMEE_FILEWRITEERROR;
						DEBUGWARNING(smee);
						smee=SMEE_OK;
						ste=STE_NORM;
					}
				}

				if (ste==STE_FILE)
				{
					if ((smee=SMNormToFile(nfh,dwOldSize,(LPSTR)lpahs,
						npts->wDataSize,npts->wSelBuffer))!=SMEE_OK)
					{
						/* Failed, so truncate back to original size... */
						DWORD dwPos;
						DEBUGWARNING(smee);

						if ((smee=FileSeek(nfh,dwOldSize,0,&dwPos))!=SMEE_OK)
						{
							DEBUGWARNING(smee);
							smee=SMEE_OK;
						}
						else
						{
							if (dwPos!=dwOldSize)
							{
								smee=SMEE_FILESEEKERROR;
								DEBUGWARNING(smee);
								smee=SMEE_OK;
							}
							else
							{
								WORD wNumWritten;

								if ((smee=FileWrite(nfh,NULL,0,&wNumWritten))!=
									SMEE_OK)
								{
									DEBUGWARNING(smee);
									smee=SMEE_OK;
								}
								else
								{
									if (wNumWritten!=0)
									{
										smee=SMEE_FILEWRITEERROR;
										DEBUGWARNING(smee);
										smee=SMEE_OK;
									}
								}
							}
						}
						ste=STE_NORM;
					}
					else
					{
						npbs->wSwapPos=wSwapPos;
						npbs->ste=STE_FILE;
						npts->wFileNum++;
					}
				}
#ifndef ALWAYSOPEN
				FileClose(nfh);
				nfh=-1;
#endif
			}

#ifdef ALWAYSOPEN
			if (wCurrentPSP!=wOwnerPSP)
			{
				/* Restore to old PSP... */
				_asm mov bx,wCurrentPSP
				_asm mov ah,050h
				DOS3Call();
			}
#endif
		}
		LocalUnlock(lhts);
		npts=NULL;
		npbs=NULL;
		GlobalUnlock(gh);
		lpahs=NULL;
	}
#ifdef DEBUG
	OutputDebugString("NotifyProc exited.\r\n");

	wReentry--;
#endif

	return ste!=STE_NORM;
}

static SMERRORENUM SMEMSBlockFree(NPTYPESTRUCT npts,WORD wbsIndex)
{
	/* Free EMS entry... */
	SMERRORENUM smee=SMEE_OK;
	WORD wEMSOldPages=(WORD)
		(((DWORD)npts->wDataSize*npts->wEMSNum+16383UL)/16384UL);
	WORD wEMSNewPages=(WORD)
		(((DWORD)npts->wDataSize*(npts->wEMSNum-1)+16383UL)/16384UL);
	NPBLOCKSTRUCT npbs=npts->abs;

	if (npts->wEMSNum>1 && npbs[wbsIndex].wSwapPos!=npts->wEMSNum-1)
	{
		/* Not at the end, so move last block down into freed pos... */
		ARENAHEADERSTRUCT ahsLast;
		WORD wSwapPos=npbs[wbsIndex].wSwapPos;
		DWORD dwPosThis=(DWORD)npts->wDataSize*wSwapPos;
		DWORD dwPosLast=(DWORD)npts->wDataSize*(npts->wEMSNum-1);

		/* First get info for last item... */
		if ((smee=SMEMSToNorm((LPSTR)&ahsLast,npts->wEMSHandle,dwPosLast,
			ARENAHEADERSTRUCTSIZE))!=SMEE_OK)
		{
			DEBUGERROR(smee);
		}

		if (smee==SMEE_OK)
		{
			/* Move last data to freed position. */
			if ((smee=SMEMSToEMS(npts->wEMSHandle,dwPosThis,dwPosLast,
				npts->wDataSize))!=SMEE_OK)
			{
				DEBUGERROR(smee);
			}
		}

		if (smee==SMEE_OK)
		{
			/* Indicate change of position of block. */
			npts->abs[ahsLast.wbsIndex].wSwapPos=wSwapPos;
		}
	}

	if (smee==SMEE_OK)
	{
		npts->wEMSNum--;

		/* Change in number of pages to allocate? */
		if (wEMSOldPages!=wEMSNewPages)
		{
			if (wEMSNewPages==0)
			{
				/* Release all... */
				if ((smee=EMSReleaseHandleAndMemory(npts->wEMSHandle))!=
					SMEE_OK)
				{
					DEBUGERROR(smee);
				}
				else
				{
					npts->wEMSHandle=NULLEMSHANDLE;
				}
			}
			else
			{
				/* Only reallocate... */
				if ((smee=EMSReallocateHandlePages(npts->wEMSHandle,
					wEMSNewPages))!=SMEE_OK)
				{
					DEBUGERROR(smee);
				}
			}
		}
	}
	return smee;
}

static SMERRORENUM SMFileBlockFree(int nfh,NPTYPESTRUCT npts,WORD wbsIndex)
{
	/* Free file block entry... */
	SMERRORENUM smee=SMEE_OK;
	DWORD dwPosLast=(DWORD)npts->wDataSize*(npts->wFileNum-1);
	NPBLOCKSTRUCT npbs=npts->abs;

	if (npts->wFileNum>1 && npbs[wbsIndex].wSwapPos!=npts->wFileNum-1)
	{
		/* Not at the end, so move last block down into freed pos... */
		ARENAHEADERSTRUCT ahsLast;
		WORD wSwapPos=npbs[wbsIndex].wSwapPos;
		DWORD dwPosThis=(DWORD)npts->wDataSize*wSwapPos;

		/* First get info for last item... */
		if ((smee=SMFileToNorm((LPSTR)&ahsLast,nfh,dwPosLast,
			ARENAHEADERSTRUCTSIZE,npts->wSelBuffer))!=SMEE_OK)
		{
			DEBUGERROR(smee);
		}

		if (smee==SMEE_OK)
		{
			/* Move last data to freed position. */
			if ((smee=SMFileToFile(nfh,dwPosThis,dwPosLast,
				npts->wDataSize,npts->wSelBuffer))!=SMEE_OK)
			{
				DEBUGERROR(smee);
			}
		}

		if (smee==SMEE_OK)
		{
			/* Indicate change of position. */
			npts->abs[ahsLast.wbsIndex].wSwapPos=wSwapPos;
		}
	}

	/* Truncate... */
	if (smee==SMEE_OK)
	{
		DWORD dwPos=0UL;

		if ((smee=FileSeek(nfh,dwPosLast,0,&dwPos))!=SMEE_OK)
		{
			DEBUGWARNING(smee);
			smee=SMEE_OK;
		}
		else
		{
			if (dwPos!=dwPosLast)
			{
				smee=SMEE_FILESEEKERROR;
				DEBUGWARNING(smee);
				smee=SMEE_OK;
			}
			else
			{
				WORD wNumWritten=0U;

				if ((smee=FileWrite(nfh,NULL,0U,&wNumWritten))!=SMEE_OK)
				{
					DEBUGWARNING(smee);
					smee=SMEE_OK;
				}
				else
				{
					if (wNumWritten!=0U)
					{
						smee=SMEE_FILEWRITEERROR;
						DEBUGWARNING(smee);
						smee=SMEE_OK;
					}
				}
			}
		}
	}

	if (smee==SMEE_OK)
	{
		npts->wFileNum--;
	}
	return smee;
}

static SMERRORENUM SMBlockUndiscard(NPTYPESTRUCT npts,WORD wbsIndex)
{
	/* Reload block from EMS. */
	SMERRORENUM smee=SMEE_OK;
	NPBLOCKSTRUCT npbs=npts->abs+wbsIndex;

#ifdef DEBUG
	OutputDebugString("Undiscard entered...\r\n");
#endif

	/* Tell NotifyProc NOT to discard this handle... */
	_ghNoDiscard=npbs->ghData;

	/* Only specify GMEM_NOTIFY here: it's the only way to allow handle */
	/* to notify again. */
	if (GlobalReAlloc(npbs->ghData,(DWORD)npts->wDataSize,GMEM_NOTIFY)==NULL)
	{
		smee=SMEE_NOMEMORY;
		DEBUGWARNING(smee);
	}
	else
	{
		LPSTR lps=GlobalLock(npbs->ghData);

		switch (npbs->ste)
		{
			case STE_EMS:
				{
					if ((smee=SMEMSToNorm(lps,npts->wEMSHandle,
						(DWORD)npbs->wSwapPos*npts->wDataSize,
						npts->wDataSize))!=SMEE_OK)
					{
						DEBUGERROR(smee);
					}
					else
					{
						npbs->ste=STE_NORM;
						if ((smee=SMEMSBlockFree(npts,wbsIndex))!=SMEE_OK)
						{
							DEBUGERROR(smee);
						}
					}
				}
				break;

			case STE_FILE:
				{
#ifdef ALWAYSOPEN
					int nfh=npts->nfh;
#else
					int nfh=-1;

					if ((smee=FileOpen(npts->szPathName,
						OF_READWRITE | OF_SHARE_EXCLUSIVE,&nfh))!=SMEE_OK)
					{
						DEBUGERROR(smee);
					}
					else
#endif
					{
						if ((smee=SMFileToNorm(lps,nfh,
							(DWORD)npbs->wSwapPos*npts->wDataSize,
							npts->wDataSize,npts->wSelBuffer))!=SMEE_OK)
						{
							DEBUGERROR(smee);
						}
						else
						{
							npbs->ste=STE_NORM;
							if ((smee=SMFileBlockFree(nfh,npts,wbsIndex))!=
								SMEE_OK)
							{
								DEBUGERROR(smee);
							}
						}
#ifndef ALWAYSOPEN
						FileClose(nfh);
						nfh=-1;
#endif
					}
				}
				break;
		}
		GlobalUnlock(npbs->ghData);
		lps=NULL;
	}

	/* Tell NotifyProc it's OK to discard now. */
	_ghNoDiscard=NULL;

#ifdef DEBUG
	OutputDebugString("Undiscard exited...\r\n");
#endif

	return smee;
}

/************************************************************************/
/*                                                                      */
/* Public functions                                                     */
/*                                                                      */
/************************************************************************/

SMERRORENUM FAR PASCAL SMTypeAlloc(WORD wDataSize,WORD wfsMin,WORD wFlags,
	LPLOCALHANDLE lplhts)
{
	/* Called by a process wishing to use a new type. */
	/* Returns an error code. */
	/* wDataSize is the size in bytes of each data block. */
	/* wfsMin is the minimum number of entries for each block's free list. */
	/* wFlags are the SMEM_XXXX flags for swapping. */
	/* *lplhts is the destination of the global handle. */
	SMERRORENUM smee=SMEE_OK;
	LOCALHANDLE lhtsResult=NULL;
	WORD wSelBuffer=NULL;

	/* Check that parameters are sensible */
	if ((((wFlags & (SMEM_SWAPFILE | SMEM_SWAPEMS))!=0)== /* Check flags */
		((wFlags & SMEM_NOSWAP)!=0)) ||
		(wFlags & ~(SMEM_NOSWAP | SMEM_SWAPFILE | SMEM_SWAPEMS))!=0 ||
		wfsMin<10 ||
		wDataSize<wfsMin*FREESTRUCTSIZE+ARENAHEADERSTRUCTSIZE+128 ||
		((_dwWinFlags & WF_STANDARD)!=0UL && wDataSize>(WORD)MAXSTANDARD))
	{
		smee=SMEE_INVALIDPARAMETER;
	}

	/* ts needs to be local: this is because we need access to this data */
	/* during NotifyProc when our app is not mapped in. Using */
	/* GMEM_NOT_BANKED didn't work: Windows gave an invalid global handle */
	/* when trying to lock the ts. */
	if (smee==SMEE_OK &&
		(lhtsResult=LocalAlloc(LMEM_MOVEABLE,TYPESTRUCTSIZE))==NULL)
	{
		return SMEE_NOMEMORY;
	}

	if (smee==SMEE_OK)
	{
		/* Get some DOS memory if we're in prot mode & we're swapping */
		/* to disk. */
		if ((_dwWinFlags & WF_PMODE)!=0 && (wFlags & SMEM_SWAPFILE)!=0 &&
			(wSelBuffer=LOWORD(GlobalDosAlloc((DWORD)BUFFERSIZE)))==NULL)
		{
			smee=SMEE_NOMEMORY;
		}

		if (smee==SMEE_OK)
		{
			/* Set ts's data... */
			NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhtsResult);

			/* Try to create file first. Can't create in NotifyProc because */
			/* GetTempFileName leads to possible heap movement. */
			if ((wFlags & SMEM_SWAPFILE)!=0)
			{
				/* Create the temporary file... */
				SMERRORENUM smee=SMEE_OK;
				int nfh=-1;

				GetTempFileName('\0',"SM_",0,npts->szPathName);

				/* Check it's there... */
				if ((smee=FileOpen(npts->szPathName,
					OF_READWRITE | OF_SHARE_EXCLUSIVE,&nfh))!=SMEE_OK)
				{
					npts->szPathName[0]='\0';
					DEBUGWARNING(smee);
				}
				else
				{
#ifdef ALWAYSOPEN
					npts->nfh=nfh;
#else
					FileClose(nfh);
#endif
				}
			}

			if (smee==SMEE_OK)
			{
				/* Now initialise vars... */
				npts->wbsNum=0;
				npts->wDataSize=wDataSize;
				npts->wfsMin=wfsMin;
				npts->wbsFreeIndex=NULLBLOCKSTRUCTINDEX; /* No free bs's. */
				npts->wEMSHandle=NULLEMSHANDLE;
				npts->wEMSNum=0;
				npts->wEMSbsIndex=NULLBLOCKSTRUCTINDEX;
				npts->wFlags=wFlags;
				npts->wFilebsIndex=NULLBLOCKSTRUCTINDEX;
				npts->wFileNum=0;
				npts->wSelBuffer=wSelBuffer;
				npts->wbsLastAllocIndex=NULLBLOCKSTRUCTINDEX;
#ifdef ALWAYSOPEN
				npts->wPSP=GetCurrentPDB();
#endif
			}
			LocalUnlock(lhtsResult);
			npts=NULL;
		}
	}

	if (smee==SMEE_OK)
	{
		LocalShrink(NULL,0);
		*lplhts=lhtsResult;
	}
	else
	{
		if (lhtsResult!=NULL)
		{
			LocalFree(lhtsResult);
			lhtsResult=NULL;
		}
		if (wSelBuffer!=NULL)
		{
			GlobalDosFree(wSelBuffer);
			wSelBuffer=NULL;
		}
	}
	return smee;
}

SMERRORENUM FAR PASCAL SMTypeFree(LOCALHANDLE lhts)
{
	/* Free all allocations for lhts. */
	SMERRORENUM smee=SMEE_OK;
	NPTYPESTRUCT npts=NULL;

	if ((npts=(NPTYPESTRUCT)LocalLock(lhts))==NULL)
	{
		smee=SMEE_INVALIDPARAMETER;
	}

	if (smee==SMEE_OK)
	{
		NPBLOCKSTRUCT npbs=npts->abs;
		WORD w;

		for (w=0;w<npts->wbsNum;w++)
		{
			if (npbs->ghData!=NULL)
			{
				if (npbs->Union.Used.wDataSeg!=NULL)
				{
					GlobalUnlock(npbs->ghData);
					npbs->Union.Used.wDataSeg=NULL;
				}
				GlobalFree(npbs->ghData);
			}
			npbs++;
		}

		/* Release EMS... */
		if (npts->wEMSHandle!=NULLEMSHANDLE)
		{
			EMSReleaseHandleAndMemory(npts->wEMSHandle);
			npts->wEMSHandle=NULLEMSHANDLE;
		}

		/* Delete any file created... */
		if ((npts->wFlags & SMEM_SWAPFILE)!=0 && npts->szPathName!='\0')
		{
#ifdef ALWAYSOPEN
			FileClose(npts->nfh);
			npts->nfh=-1;
#endif
			FileDelete(npts->szPathName);
		}

		/* Delete any Dos memory created... */
		if (npts->wSelBuffer!=NULL)
		{
			GlobalDosFree(npts->wSelBuffer);
		}

		LocalUnlock(lhts);
		npbs=NULL;
		npts=NULL;
		LocalFree(lhts);
		LocalShrink(NULL,0);
	}
	return smee;
}

SMERRORENUM FAR PASCAL SMLock(LOCALHANDLE lhts,SMHANDLE smh,LPLPVOID lplp)
{
	/* Lock lhts/smh. */
	SMERRORENUM smee=SMEE_OK;
	WORD wbsIndex=HIWORD((DWORD)smh);
	NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhts);
	LPVOID lpResult=NULL;
	NPBLOCKSTRUCT npbs=npts->abs+wbsIndex;

	if (wbsIndex>=npts->wbsNum || npbs->ghData==NULL)
	{
		smee=SMEE_INVALIDPARAMETER;
	}
	else
	{
		if (npbs->Union.Used.wDataSeg==NULL)
		{
			if (npbs->ste!=STE_NORM)
			{
				/* Reload from swap info... */
				smee=SMBlockUndiscard(npts,wbsIndex);
			}

			if (smee==SMEE_OK)
			{
				npbs->Union.Used.wDataSeg=
					HIWORD((DWORD)GlobalLock(npbs->ghData));
			}
		}
		npbs->wLockCount++;
		lpResult=(LPVOID)MAKELONG(LOWORD(smh),npbs->Union.Used.wDataSeg);
	}
	LocalUnlock(lhts);
	npbs=NULL;
	npts=NULL;
	if (smee==SMEE_OK)
	{
		*lplp=lpResult;
	}
	return smee;
}

SMERRORENUM FAR PASCAL SMUnlock(LOCALHANDLE lhts,SMHANDLE smh)
{
	/* Unlock lhts/smh. */
	SMERRORENUM smee=SMEE_OK;
	WORD wbsIndex=HIWORD((DWORD)smh);
	NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhts);
	NPBLOCKSTRUCT npbs=npts->abs+wbsIndex;

	if (wbsIndex>=npts->wbsNum || npbs->ghData==NULL)
	{
		smee=SMEE_INVALIDPARAMETER;
	}
	else
	{
		npbs->wLockCount--;
		if (npbs->wLockCount==0)
		{
			GlobalUnlock(npbs->ghData);
			npbs->Union.Used.wDataSeg=NULL;
		}
	}
	LocalUnlock(lhts);
	npbs=NULL;
	npts=NULL;
	return smee;
}

static SMERRORENUM SMBlockAlloc(LOCALHANDLE lhts,LPWORD lpwbsIndex)
{
	/* Allocate & initialise a new block to lhts. Returns possible
	/* error. */
	/* lhts should be unlocked, as a realloc may occur here. */
	/* *lpwbsIndex is the bs index to use. */
	SMERRORENUM smee=SMEE_OK;
	GLOBALHANDLE ghData;
	WORD wbsIndex=0;
	NPBLOCKSTRUCT npbs=NULL;
	NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhts);

	if ((ghData=GlobalAlloc(GMEM_MOVEABLE |
		((npts->wFlags & (SMEM_SWAPFILE | SMEM_SWAPEMS))!=0 ?
		GMEM_DISCARDABLE | GMEM_NOTIFY : 0),(DWORD)npts->wDataSize))==NULL)
	{
		smee=SMEE_NOMEMORY;
	}
	else
	{
		/* Tell NotifyProc not to discard ghData... */
		_ghNoDiscard=ghData;

		/* Try to locate a spare bs. if wbsFreeIndex!=NULLBLOCKSTRUCTINDEX, */
		/* then there's a free. */
		if (npts->wbsFreeIndex!=NULLBLOCKSTRUCTINDEX)
		{
			NPBLOCKSTRUCT npbsBase=npts->abs;
			WORD wbsFreePrev;
			WORD wbsFreeNext;

			wbsIndex=npts->wbsFreeIndex;
			npbs=npts->abs+wbsIndex;
			wbsFreeNext=npbs->Union.Free.wbsFreeNext;
			wbsFreePrev=npbs->Union.Free.wbsFreePrev;

			/* The only free bs? */
			if (wbsFreeNext!=wbsIndex)
			{
				/* No, so remove from linked list... */
				npts->wbsFreeIndex=wbsFreeNext;
				npbsBase[wbsFreePrev].Union.Free.wbsFreeNext=wbsFreeNext;
				npbsBase[wbsFreeNext].Union.Free.wbsFreePrev=wbsFreePrev;
			}
			else
			{
				/* Yes, this is the only free bs, so nullify ts entry index. */
				npts->wbsFreeIndex=NULLBLOCKSTRUCTINDEX;
			}
		}
		else
		{
			/* Need to reallocate, as no spare bs... */
			DWORD dwSize=
				(DWORD)BLOCKSTRUCTSIZE*(npts->wbsNum+1)+TYPESTRUCTSIZE;

			/* Unlock for realloc... */
			LocalUnlock(lhts);
			npts=NULL;
			npbs=NULL;

			/* Check the size isn't bigger than 64K. */
			if (dwSize>=65536UL)
			{
				smee=SMEE_TOOMANYBLOCKS;
			}
			if (smee==SMEE_OK && LocalReAlloc(lhts,(WORD)dwSize,0)==NULL)
			{
				smee=SMEE_NOMEMORY;
			}
			if (smee==SMEE_OK)
			{
				/* Lock again... */
				npts=(NPTYPESTRUCT)LocalLock(lhts);

				wbsIndex=npts->wbsNum;
				npts->wbsNum++;
				npbs=npts->abs+wbsIndex;
			}
		}

		if (smee==SMEE_OK)
		{
			npbs->ghData=ghData;
			npbs->Union.Used.wDataSeg=NULL;
			npbs->Union.Used.wDataNum=0U;
			npbs->wfsNum=1;
			npbs->wfsLargestIndex=0U;

			/* Make LargestSize so that we've enough space for min free */
			/* entries. */
			npbs->wfsLargestSize=npts->wDataSize-FREESTRUCTSIZE*npts->wfsMin-
				ARENAHEADERSTRUCTSIZE;
			npbs->wLockCount=0U;
			npbs->wSwapPos=0U;
			npbs->ste=STE_NORM;

			/* Set up the free list... */
			{
				LPFREESTRUCT lpfs=(LPFREESTRUCT)(GlobalLock(ghData)+
					npts->wDataSize)-npbs->wfsNum;

				lpfs->wOffStart=ARENAHEADERSTRUCTSIZE;
				lpfs->wOffEnd=npts->wDataSize-FREESTRUCTSIZE*npts->wfsMin;
				GlobalUnlock(ghData);
				lpfs=NULL;
			}

			/* Set up the arena header... */
			{
				LPARENAHEADERSTRUCT lpahs=
					(LPARENAHEADERSTRUCT)GlobalLock(ghData);

				lpahs->lhts=lhts;
				lpahs->wbsIndex=wbsIndex;
				GlobalUnlock(ghData);
				lpahs=NULL;
			}
			/* Indicate that this is the latest block alllocated. */
			npts->wbsLastAllocIndex=wbsIndex;
		}

		/* Tell NotifyProc it's OK to discard when unlocked again. */
		_ghNoDiscard=NULL;
	}
	if (npts!=NULL)
	{
		LocalUnlock(lhts);
		npts=NULL;
		npbs=NULL;
	}

	if (smee==SMEE_OK)
	{
		*lpwbsIndex=wbsIndex;
	}
	else
	{
		if (ghData!=NULL)
		{
			GlobalFree(ghData);
		}
	}
	return smee;
}

static SMERRORENUM SMBlockFree(LOCALHANDLE lhts,WORD wbsIndex)
{
	/* Free a block of lhts, index wbsIndex. */
	/* lhts should be unlocked when called to allow possible realloc. */
	NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhts);
	NPBLOCKSTRUCT npbsBase=npts->abs;
	NPBLOCKSTRUCT npbs=npbsBase+wbsIndex;
	WORD wbsNum=npts->wbsNum;

	switch (npbs->ste)
	{
		case STE_NORM:
			if (npbs->Union.Used.wDataSeg!=NULL)
			{
				GlobalUnlock(npbs->ghData);
				npbs->Union.Used.wDataSeg=NULL;
			}
			break;

		case STE_EMS:
			SMEMSBlockFree(npts,wbsIndex);
			break;

		case STE_FILE:
			{
#ifdef ALWAYSOPEN
				int nfh=npts->nfh;
#else
				int nfh=-1;

				if (FileOpen(npts->szPathName,
					OF_READWRITE | OF_SHARE_EXCLUSIVE,&nfh)==SMEE_OK)
#endif
				{
					SMFileBlockFree(nfh,npts,wbsIndex);
#ifndef ALWAYSOPEN
					FileClose(nfh);
					nfh=-1;
#endif
				}
			}
			break;
	}
	GlobalFree(npbs->ghData);
	npbs->ghData=NULL;

	/* ReAlloc bs if we're the last entry... */
	if (wbsIndex==wbsNum-1)
	{
		npbs--;
		wbsNum--;

		/* Find any additional trailing entries... */
		while (wbsNum>0 && npbs->ghData==NULL)
		{
			/* Remove from bs free linked list... */
			WORD wbsFreeNext=npbs->Union.Free.wbsFreeNext;
			WORD wbsFreePrev=npbs->Union.Free.wbsFreePrev;

			if (wbsFreeNext!=wbsNum-1)
			{
				/* Not the last entry in free list. */
				npbsBase[wbsFreePrev].Union.Free.wbsFreeNext=wbsFreeNext;
				npbsBase[wbsFreeNext].Union.Free.wbsFreePrev=wbsFreePrev;

				if (npts->wbsFreeIndex==wbsNum)
				{
					npts->wbsFreeIndex=wbsFreeNext;
				}
			}
			else
			{
				/* Last entry in free list, so clear ts's FreeIndex. */
				npts->wbsFreeIndex=NULLBLOCKSTRUCTINDEX;
			}
			npbs--;
			wbsNum--;
		}
		npts->wbsNum=wbsNum;
	}
	else
	{
		/* Insert into bs linked list... */
		if (npts->wbsFreeIndex!=NULLBLOCKSTRUCTINDEX)
		{
			WORD wbsFreeNext=npts->wbsFreeIndex;
			WORD wbsFreePrev=npbsBase[wbsFreeNext].Union.Free.wbsFreePrev;

			npbs->Union.Free.wbsFreeNext=wbsFreeNext;
			npbs->Union.Free.wbsFreePrev=wbsFreePrev;
			npbsBase[wbsFreeNext].Union.Free.wbsFreePrev=wbsIndex;
			npbsBase[wbsFreePrev].Union.Free.wbsFreeNext=wbsIndex;
		}
		else
		{
			npbs->Union.Free.wbsFreeNext=wbsIndex;
			npbs->Union.Free.wbsFreePrev=wbsIndex;
			npts->wbsFreeIndex=wbsIndex;
		}
	}
	LocalUnlock(lhts);
	npts=NULL;
	npbsBase=NULL;
	npbs=NULL;
	if (wbsIndex>=wbsNum)
	{
		LocalReAlloc(lhts,BLOCKSTRUCTSIZE*wbsNum+TYPESTRUCTSIZE,0);
		LocalShrink(NULL,0);
	}
	return SMEE_OK;
}

static VOID SMSetLargestFree(NPBLOCKSTRUCT npbs,LPFREESTRUCT lpfs)
{
	/* Locate biggest free space now... */
	/* lpfs points to the BASE of the free list on entry. */
	WORD w;

	npbs->wfsLargestSize=0;

	for (w=0;w<npbs->wfsNum;w++)
	{
		WORD wSize=lpfs->wOffEnd-lpfs->wOffStart;

		if (wSize>npbs->wfsLargestSize)
		{
			npbs->wfsLargestSize=wSize;
			npbs->wfsLargestIndex=w;
		}
		lpfs++;
	}
}

SMERRORENUM FAR PASCAL SMAlloc(LOCALHANDLE lhts,WORD wSize,
	LPSMHANDLE lpsmh)
{
	/* Return an smh to memory allocated to lhts. */
	SMERRORENUM smee=SMEE_OK;
	WORD wbsIndex=0;
	NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhts);
	NPBLOCKSTRUCT npbs=npts->abs;
	WORD wOffset=0;
	WORD wbsNum=npts->wbsNum;

	if (wSize>npts->wDataSize-npts->wfsMin*FREESTRUCTSIZE-
		ARENAHEADERSTRUCTSIZE)
	{
		smee=SMEE_SIZETOOLARGE;
	}
	else
	{
		/* Try to locate a bs with enough space... */

		/* Speed enhancement trick: try the last allocated bs first... */
		if (npts->wbsLastAllocIndex<wbsNum &&
			npbs[npts->wbsLastAllocIndex].ghData!=NULL &&
			npbs[npts->wbsLastAllocIndex].wfsLargestSize>=wSize)
		{
			wbsIndex=npts->wbsLastAllocIndex;
		}
		else
		{
			/* Otherwise do a search for the space. */
			while (wbsIndex<wbsNum &&
				(npbs->ghData==NULL || npbs->wfsLargestSize<wSize))
			{
				wbsIndex++;
				npbs++;
			}
		}
		/* Unlock for SMBlockAlloc call... */
		LocalUnlock(lhts);
		npts=NULL;
		npbs=NULL;

		if (wbsIndex>=wbsNum)
		{
			/* Try to get a new allocation. */
			smee=SMBlockAlloc(lhts,&wbsIndex);
		}
		if (smee==SMEE_OK)
		{
			/* We've got memory to do this allocation, so do it... */
			npts=(NPTYPESTRUCT)LocalLock(lhts);
			npbs=npts->abs+wbsIndex;

			if (npbs->ste!=STE_NORM)
			{
				/* Reload if necessary... */
				smee=SMBlockUndiscard(npts,wbsIndex);
			}

			if (smee==SMEE_OK)
			{
				/* Tell NotifyProc not to discard me... */
				_ghNoDiscard=npbs->ghData;

				/* Modify free list... */
				if (wSize==0)
				{
					/* Special case of zero allocation size. */
					wOffset=ARENAHEADERSTRUCTSIZE;
				}
				else
				{
					if (npbs->wfsLargestSize==wSize)
					{
						/* Remove free list entry entirely as free list entry */
						/* same size as allocation. */
						WORD w=npbs->wfsLargestIndex;
						LPFREESTRUCT lpfsBase=
							(LPFREESTRUCT)(GlobalLock(npbs->ghData)+
							npts->wDataSize)-npbs->wfsNum;
						LPFREESTRUCT lpfs=lpfsBase+w;

						wOffset=lpfs->wOffStart;

						/* Move previous free list entries up, deleting this */
						/* one... */
						for (;w>0;w--)
						{
							lpfs[0]=lpfs[-1];
							lpfs--;
						}
						lpfsBase++; /* Update to new fs base position. */
						npbs->wfsNum--; /* One less fs. */

						/* Possible smaller realloc of fs? */
						if (npbs->wfsNum>=npts->wfsMin)
						{
							lpfsBase[npbs->wfsNum-1].wOffEnd+=FREESTRUCTSIZE;
						}
						GlobalUnlock(npbs->ghData);
						lpfsBase=NULL;
						lpfs=NULL;
					}
					else
					{
						/* Modify free list entry. */
						LPFREESTRUCT lpfs=(LPFREESTRUCT)(GlobalLock(npbs->ghData)+
							npts->wDataSize)-npbs->wfsNum+npbs->wfsLargestIndex;

						wOffset=lpfs->wOffStart;

						lpfs->wOffStart+=wSize;
						GlobalUnlock(npbs->ghData);
						lpfs=NULL;
					}

					/* Locate biggest free space now... */
					{
						LPFREESTRUCT lpfsBase=
							(LPFREESTRUCT)(GlobalLock(npbs->ghData)+
							npts->wDataSize)-npbs->wfsNum;

						SMSetLargestFree(npbs,lpfsBase);
						GlobalUnlock(npbs->ghData);
						lpfsBase=NULL;
					}
				}
				/* Tell NotifyProc it's OK to discard me... */
				_ghNoDiscard=NULL;

				npbs->Union.Used.wDataNum++;
			}
		}
	}
	if (npts!=NULL)
	{
		LocalUnlock(lhts);
		npts=NULL;
		npbs=NULL;
	}

	if (smee==SMEE_OK)
	{
		*lpsmh=MAKELONG(wOffset,wbsIndex);
	}
	return smee;
}

SMERRORENUM FAR PASCAL SMFree(LOCALHANDLE lhts,SMHANDLE smh,WORD wSize)
{
	/* Free an smh... */
	SMERRORENUM smee=SMEE_OK;
	WORD wbsIndex=HIWORD((DWORD)smh);
	WORD wOffset=LOWORD((DWORD)smh);
	NPTYPESTRUCT npts=(NPTYPESTRUCT)LocalLock(lhts);
	NPBLOCKSTRUCT npbs=npts->abs+wbsIndex;
	WORD wDataNum=0; /* Initialising prevents MSC 6 warning. */

	if (wbsIndex>=npts->wbsNum || npbs->ghData==NULL)
	{
		smee=SMEE_INVALIDPARAMETER;
	}
	else
	{
		/* Special case of zero size... */
		if (wSize!=0)
		{
			if (npbs->ste!=STE_NORM)
			{
				/* Reload if necessary... */
				smee=SMBlockUndiscard(npts,wbsIndex);
			}

			if (smee==SMEE_OK)
			{
				/* Tell NotifyProc not to discard me... */
				_ghNoDiscard=npbs->ghData;

				/* First see if we need to enlarge the free list... */
				{
					LPFREESTRUCT lpfsBase=
						(LPFREESTRUCT)(GlobalLock(npbs->ghData)+
						npts->wDataSize)-npbs->wfsNum;
					LPFREESTRUCT lpfs=lpfsBase;
					BOOL bPrevContig=FALSE;
					BOOL bNextContig=FALSE;
					WORD wfsIndex=0;

					/* Will this make a contiguous free block? */
					/******** Maybe a binary search would be nice here **********/
					while (wfsIndex<npbs->wfsNum && lpfs->wOffStart<wOffset)
					{
						wfsIndex++;
						lpfs++;
					}

					if (wfsIndex>0 && lpfs[-1].wOffEnd==wOffset)
					{
						/* The previous entry is contiguous with the one to */
						/* delete. */
						lpfs[-1].wOffEnd+=wSize;
						bPrevContig=TRUE;
					}
					if (wfsIndex<npbs->wfsNum &&
						lpfs->wOffStart==wOffset+wSize)
					{
						/* The next entry is contiguous with the one to */
						/* delete. */
						if (bPrevContig)
						{
							WORD w;

							/* Combine... */
							lpfs[-1].wOffEnd=lpfs->wOffEnd;

							/* Delete entry as we've combined three free */
							/* entries... */
							for (w=wfsIndex;w>0;w--)
							{
								lpfs[0]=lpfs[-1];
								lpfs--;
							}
							lpfsBase++;
							npbs->wfsNum--;

							/* Possible realloc decrease of fs? */
							if (npbs->wfsNum>=npts->wfsMin)
							{
								lpfsBase[npbs->wfsNum-1].wOffEnd+=
									FREESTRUCTSIZE;
							}
						}
						else
						{
							lpfs->wOffStart=wOffset;
						}
						bNextContig=TRUE;
					}

					/* Do we need to insert a new fs entry? */
					if (!bPrevContig && !bNextContig)
					{
						/* Does fs need a bigger size? */
						if (npbs->wfsNum+1>npts->wfsMin)
						{
							/* Yes, but can we do it? */
							lpfs=lpfsBase+npbs->wfsNum-1;
							if (lpfs->wOffEnd!=
								npts->wDataSize-FREESTRUCTSIZE*npbs->wfsNum ||
								lpfs->wOffEnd-lpfs->wOffStart<FREESTRUCTSIZE)
							{
								smee=SMEE_FREEEXHAUSTED;
							}
							else
							{
								lpfs->wOffEnd-=FREESTRUCTSIZE;
							}
						}

						if (smee==SMEE_OK)
						{
							/* Now to insert data... */
							lpfsBase--; /* New base position. */
							{
								WORD w;

								lpfs=lpfsBase;

									for (w=0;w<wfsIndex;w++)
								{
									lpfs[0]=lpfs[1];
									lpfs++;
								}

								lpfs->wOffStart=wOffset;
								lpfs->wOffEnd=wOffset+wSize;
								npbs->wfsNum++;
							}
						}
					}

					if (smee==SMEE_OK)
					{
						SMSetLargestFree(npbs,lpfsBase);
					}

					GlobalUnlock(npbs->ghData);
					lpfsBase=NULL;
					lpfs=NULL;
				}
				/* Tell NotifyProc it's OK to discard me... */
				_ghNoDiscard=NULL;

				if (smee==SMEE_OK)
				{
					npbs->Union.Used.wDataNum--;
					wDataNum=npbs->Union.Used.wDataNum;
				}
			}
		}
	}
	LocalUnlock(lhts);
	npts=NULL;
	npbs=NULL;

	if (smee==SMEE_OK && wDataNum==0)
	{
		SMBlockFree(lhts,wbsIndex);
	}
	return smee;
}

VOID FAR PASCAL SMInit(void)
{
	GlobalNotify(NotifyProc);
}
