#define NOCOMM
#include <windows.h>

#include <stdlib.h>
#include <string.h>

#include "smalloc.h"

#define NUM_HANDLES 7000
#define MAX_SIZE 1024
#define BLOCK_SIZE 9999
#define FIELD_WIDTH 14
#define NUM_FREE 32

#define IDM_ALLOC 101
#define IDM_LOCK 102
#define IDM_FREE 103

typedef struct HandleInfoStruct
{
	SMHANDLE smh;
	WORD wSize;
	BOOL bReShow;		/* Set to TRUE if there's been an obvious change. */
} HandleInfoStruct;

static HandleInfoStruct ahis[NUM_HANDLES];

typedef struct
{
	SMHANDLE smh;
	char szInfo[30];
} InfoStruct;

static BOOL bAlloc=FALSE;
static BOOL bLock=FALSE;
static BOOL bFree=FALSE;

static TEXTMETRIC tm;

static char szAppName[]="TestVirt";

static char szTestString[]="Hello from Howie.";

static GLOBALHANDLE ghtu=NULL;

static void DoPaint(HDC hDC,BOOL bShowAll)
{
	int n;

	for (n=0;n<NUM_HANDLES;n++)
	{
		if (bShowAll || ahis[n].bReShow)
		{
			char szTemp[FIELD_WIDTH+1];

			ahis[n].bReShow=FALSE;

			if (ahis[n].smh==NULL)
			{
				wsprintf(szTemp,"              ");
			}
			else
			{
				wsprintf(szTemp,"%08lX %03X  ",ahis[n].smh,ahis[n].wSize);
			}
			TextOut(hDC,(n%5)*FIELD_WIDTH*tm.tmAveCharWidth,
				(n/5)*tm.tmHeight,szTemp,FIELD_WIDTH);
		}
	}
}

long FAR PASCAL WndProc(HWND hWnd,unsigned iMessage,WORD wParam,
	LONG lParam)
{
	switch (iMessage)
	{
		case WM_CREATE:
			{
				int i;

				for (i=0;i<NUM_HANDLES;i++)
				{
					ahis[i].smh=NULL;
					ahis[i].wSize=0;
					ahis[i].bReShow=FALSE;
				}
			}
			{
				HMENU hMenu=GetSystemMenu(hWnd,FALSE);

				ChangeMenu(hMenu,NULL,NULL,NULL,MF_APPEND|MF_SEPARATOR);
				ChangeMenu(hMenu,NULL,"Random allocation",IDM_ALLOC,
					MF_APPEND|MF_STRING);
				ChangeMenu(hMenu,NULL,"Random locking",IDM_LOCK,
					MF_APPEND|MF_STRING);
				ChangeMenu(hMenu,NULL,"Random freeing",IDM_FREE,
					MF_APPEND|MF_STRING);
			}
			{
				HDC hDC=GetDC(hWnd);

				GetTextMetrics(hDC,&tm);
				ReleaseDC(hWnd,hDC);
			}
			{
				SMERRORENUM smee=SMEE_OK;

				SMInit();

				if ((smee=SMTypeAlloc(BLOCK_SIZE,NUM_FREE,SMEM_SWAPEMS | SMEM_SWAPFILE,
					&ghtu))!=SMEE_OK)
				{
					char sz[128];
					wsprintf(sz,"SMTypeAlloc Error #%d",smee);
					MessageBox(hWnd,sz,szAppName,MB_OK);
					DestroyWindow(hWnd);
				}
			}
			break;

		case WM_DESTROY:
			PostQuitMessage(0); /* No break! */
		case WM_ENDSESSION:
			if (ghtu!=NULL)
			{
				SMERRORENUM smee=NULL;

				if ((smee=SMTypeFree(ghtu))!=SMEE_OK)
				{
					char sz[128];
					wsprintf(sz,"SMTypeFree Error #%d",smee);
					MessageBox(NULL,sz,szAppName,MB_OK);
				}
				ghtu=NULL;
			}
			break ;

		case WM_PAINT:
			{
				PAINTSTRUCT ps;
				HDC hDC=BeginPaint(hWnd,&ps);

				DoPaint(hDC,TRUE);

				EndPaint(hWnd,&ps);
			}
			break;

		case WM_INITMENUPOPUP:
			if (HIWORD(lParam)!=0)
			{
				CheckMenuItem(wParam,IDM_ALLOC,
					MF_BYCOMMAND|(bAlloc?MF_CHECKED:MF_UNCHECKED));
				CheckMenuItem(wParam,IDM_LOCK,
					MF_BYCOMMAND|(bLock?MF_CHECKED:MF_UNCHECKED));
				CheckMenuItem(wParam,IDM_FREE,
					MF_BYCOMMAND|(bFree?MF_CHECKED:MF_UNCHECKED));
			}
			break;

		case WM_SYSCOMMAND:
			switch (wParam)
			{
				case IDM_ALLOC:
					bAlloc=!bAlloc;
					break;
				case IDM_LOCK:
					bLock=!bLock;
					break;
				case IDM_FREE:
					bFree=!bFree;
					break;
				default:
					return DefWindowProc(hWnd,iMessage,wParam,lParam);
			}
			break;

		default:
			return DefWindowProc(hWnd,iMessage,wParam,lParam);
	}
	return 0L ;
}

void DoOurThing(HWND hWnd)
{
	if (bAlloc)
	{
		int n=rand()%NUM_HANDLES;
		WORD wSize=(WORD)(rand())%MAX_SIZE+sizeof(InfoStruct);

		if (ahis[n].smh==NULL)
		{
			SMERRORENUM smee=SMEE_OK;

			if ((smee=SMAlloc(ghtu,wSize,&ahis[n].smh))!=SMEE_OK)
			{
				char sz[128];

				wsprintf(sz,"Error %04X allocating handle",smee);
				MessageBox(NULL,sz,szAppName,MB_OK | MB_ICONASTERISK);
			}
			else
			{
				InfoStruct far *lpis=NULL;

				ahis[n].wSize=wSize;

				if ((smee=SMLock(ghtu,ahis[n].smh,(LPLPVOID)&lpis))!=SMEE_OK)
				{
					char sz[128];

					wsprintf(sz,"SMLock 1 Error #%04X @%d",smee,n);
					MessageBox(NULL,sz,szAppName,MB_OK);
				}
				if (smee==SMEE_OK)
				{
					if (lpis==NULL)
					{
						char sz[128];

						wsprintf(sz,"NULL lpis 1 Error #%04X @%d",smee,n);
						MessageBox(NULL,sz,szAppName,MB_OK);
					}
					else
					{
						lpis->smh=ahis[n].smh;
						lstrcpy(lpis->szInfo,szTestString);
						if ((smee=SMUnlock(ghtu,ahis[n].smh))!=SMEE_OK)
						{
							char sz[128];

							wsprintf(sz,"SMUnlock 1 Error #%04X @%d",smee,n);
							MessageBox(NULL,sz,szAppName,MB_OK);
						}
					}
				}
			}
			ahis[n].bReShow=TRUE;
		}
	}
	if (bLock)
	{
		int n=rand()%NUM_HANDLES;

		if (ahis[n].smh!=NULL)
		{
			InfoStruct far *lpis=NULL;
			SMERRORENUM smee=SMEE_OK;

			if ((smee=SMLock(ghtu,ahis[n].smh,(LPLPVOID)&lpis))!=SMEE_OK)
			{
				char sz[128];

				wsprintf(sz,"SMLock 2 Error #%04X @%d",smee,n);
				MessageBox(NULL,sz,szAppName,MB_OK);
			}
			if (smee==SMEE_OK)
			{
				if (lpis==NULL)
				{
					char sz[128];

					wsprintf(sz,"NULL lpis 2 Error #%04X @%d",smee,n);
					MessageBox(NULL,sz,szAppName,MB_OK);
				}
				else
				{
					if (lpis->smh!=ahis[n].smh ||
						_fstrcmp(lpis->szInfo,szTestString)!=0)
					{
						char sz[128];

						wsprintf(sz,"Trashed lpis 2 Error #%04X @%d",smee,n);
						MessageBox(NULL,sz,szAppName,MB_OK);
					}
					if ((smee=SMUnlock(ghtu,ahis[n].smh))!=SMEE_OK)
					{
						char sz[128];

						wsprintf(sz,"SMUnlock 2 Error #%04X @%d",smee,n);
						MessageBox(NULL,sz,szAppName,MB_OK);
					}
				}
			}
			ahis[n].bReShow=TRUE;
		}
	}
	if (bFree)
	{
		int n=rand()%NUM_HANDLES;

		if (ahis[n].smh!=NULL)
		{
			SMERRORENUM smee=SMEE_OK;

			if ((smee=SMFree(ghtu,ahis[n].smh,ahis[n].wSize))!=SMEE_OK)
			{
				char sz[128];

				wsprintf(sz,"SMFree 3 Error #%04X @%d",smee,n);
				MessageBox(NULL,sz,szAppName,MB_OK);
			}
			else
			{
				ahis[n].wSize=NULL;
				ahis[n].smh=NULL;
				ahis[n].bReShow=TRUE;
			}
		}
	}
	{
		HDC hDC=GetDC(hWnd);

		DoPaint(hDC,FALSE);
		ReleaseDC(hWnd,hDC);
	}
}

int PASCAL WinMain(HANDLE hInstance,HANDLE hPrevInstance,LPSTR lpszCmdLine,
	int nCmdShow)
{
	static char szAppName[]="TestVirt";

	if (!hPrevInstance)
	{
		WNDCLASS wc;

		wc.style=CS_HREDRAW|CS_VREDRAW;
		wc.lpfnWndProc=WndProc;
		wc.cbClsExtra=0;
		wc.cbWndExtra=0;
		wc.hInstance=hInstance;
		wc.hIcon=NULL;
		wc.hCursor=LoadCursor(NULL,IDC_ARROW);
		wc.hbrBackground=GetStockObject(WHITE_BRUSH);
		wc.lpszMenuName=NULL ;
		wc.lpszClassName=szAppName ;

		if (!RegisterClass(&wc))
		{
			return FALSE ;
		}
	}

	{
		HWND hWnd=CreateWindow(szAppName,"Virtual Data Test Application",
			WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,0,CW_USEDEFAULT,0,NULL,NULL,
			hInstance,NULL);

		if (hWnd==NULL)
		{
			return FALSE;
		}

		ShowWindow(hWnd,nCmdShow);
		UpdateWindow(hWnd);

		while(TRUE)
		{
			MSG msg;

			if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
			{
				if (msg.message==WM_QUIT)
				{
					return msg.wParam;
				}

				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
			else
			{
				DoOurThing(hWnd);
			}
		}
	}
}

